/*
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Copyright (C) 2007 Dr. Ruediger Berlich - ruediger@berlich.de
 */

// On OpenSUSE 10.3, compile with the "build.sh" script. 
// Tested on OpenSUSE 10.3, g++ 4.21, boost 1.34.1 . Note that, by
// default, OpenSUSE 10.3 still comes with boost 1.33.1 . In order to 
// compile Boost 1.34.1 on OpenSUSE 10.3, you might have to add the 
// compiler switch  -fno-strict-aliasing in line 155 of the file 
// $BOOSTHOME/tools/jam/src/build.jam . 
// Note that we do assume in this example that you have installed 
// both the Boost.Asio library and the threadpool library. Both 
// are not delivered by default with Boost 1.34.1 . See the second
// part of the tutorial for instructions on how to install them.


#include <cstdlib>
#include <cstring>
#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <ctime>

#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/random.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/asio.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/threadpool.hpp>

using namespace std;
using namespace boost;
using boost::asio::ip::tcp;

const unsigned int COMMANDLENGTH=32;
const unsigned int NPROCESS=100;
const unsigned int VECTORSIZE=10000;

/******************************************************************************************/

class secretNumber
{
  ///////////////////////////////////////////////////////////////
  friend class boost::serialization::access;
  
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version){
    using boost::serialization::make_nvp;
    ar & make_nvp("secret_", secret_);
  }
  ///////////////////////////////////////////////////////////////
  
public:
  secretNumber(void) : secret_(0) { /* nothing */ }
  secretNumber(double secret) : secret_(secret) { /* nothing */ }
  double value(void) const { return secret_; }

private:
  double secret_;
};

/******************************************************************************************/

template <class T>
class base
  :public vector<T>
{
  ///////////////////////////////////////////////////////////////
  friend class boost::serialization::access;

  template<class Archive>
  void serialize(Archive & ar, const unsigned int version){
    using boost::serialization::make_nvp;
    ar & make_nvp("stdvector", boost::serialization::base_object<vector<T> >(*this));
  }
  ///////////////////////////////////////////////////////////////

public:
  virtual void sort(void) = 0;
};

namespace boost {
  namespace serialization {
    template<class T>
    struct is_abstract<base<T> > {
      typedef mpl::bool_<true> type;
      BOOST_STATIC_CONSTANT(bool, value = true);
    };
}}

/******************************************************************************************/

class secretContainer
  :public base<shared_ptr<secretNumber> >
{
  ///////////////////////////////////////////////////////////////
  friend class boost::serialization::access;
  
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version){
    using boost::serialization::make_nvp;
    ar & make_nvp("base", boost::serialization::base_object<base<shared_ptr<secretNumber> > >(*this));
  }
  ///////////////////////////////////////////////////////////////

public:
  // Initialize random number generators only
  secretContainer(void)
    :mersenne(static_cast<unsigned int> (time(0))),
     normalDist(0.,1.),
     gauss(mersenne,normalDist)
  { /* nothing */ }

  // Initialize container with random numbers
  secretContainer(unsigned int size)
    :mersenne(static_cast<unsigned int> (time(0))),
     normalDist(0.,1.),
     gauss(mersenne,normalDist)
  {
    for(unsigned int i=0; i<size; i++){ 
      shared_ptr<secretNumber> p(new secretNumber(gauss()));
      this->push_back(p);
    }
  }

  string toString(void){
    ostringstream oss;  // serialize
    {
#ifdef XMLARCHIVE
      boost::archive::xml_oarchive oa(oss);
#else
      boost::archive::text_oarchive oa(oss);
#endif
      oa << boost::serialization::make_nvp("secretContainer",*this);
    } // archive and stream closed at end of scope    

    return oss.str();
  }

  void fromString(const string& descr){
    // de-serialize
    istringstream iss(descr);
    {
#ifdef XMLARCHIVE
      boost::archive::xml_iarchive ia(iss);
#else
      boost::archive::text_iarchive ia(iss);
#endif
      ia >> boost::serialization::make_nvp("secretContainer",*this);
    }
  }

  virtual void sort(void){
    std::sort(this->begin(), this->end(), 
	      bind(&secretNumber::value,_1) < bind(&secretNumber::value,_2));
  }

private:
  mt19937 mersenne;
  normal_distribution<double> normalDist;
  variate_generator<mt19937&, normal_distribution<double> > gauss;
};

/******************************************************************************************/

ostream& operator<< (ostream & s, const secretContainer& sc){
  vector<shared_ptr<secretNumber> >::const_iterator it;
  s.precision(2);
  for(it=sc.begin(); it!=sc.end(); ++it) s << fixed << (*it)->value() << " ";
  return s;
}

/******************************************************************************************/

string commandString(const string& cmd, unsigned int command_length=COMMANDLENGTH){
  std::ostringstream command_stream;
  command_stream << std::setw(command_length) << cmd;
  return command_stream.str();
}

/******************************************************************************************/

class client
{
public:
  client(string server, string port)
    :socket_(io_service_),
     resolver_(io_service_),
     query_(server, port)     
  { endpoint_iterator0=resolver_.resolve(query_); }

  void run(void) {
    for(unsigned int i=0; i<NPROCESS; i++){
      if(!tryConnect()){
	cout << "Error: Could not connect to server" << endl;
	break;
      }

      string data=retrieve();
      
      if(data == "error"){
	cout << "Error: Invalid response from server" << endl;
	break;
      }

      // De-serialize the object
      secretContainer sc;
      sc.fromString(data);

      // Sort it. This is where the actual work is done
      sc.sort();

      if(!tryConnect()){
	cout << "Error: Could not connect to server" << endl;
	break;
      }

      // And send it off again
      submit(sc.toString());

      cout << "Processed one data set" << endl;
    }
  }

private:
  bool tryConnect(void){
      boost::asio::ip::tcp::resolver::iterator endpoint_iterator=endpoint_iterator0;
      boost::system::error_code error = boost::asio::error::host_not_found;
      while (error && endpoint_iterator != end){
	socket_.close();
	socket_.connect(*endpoint_iterator++, error);
      }

      if (error) return false;
      return true;
  }

  string retrieve(void){
    boost::asio::write(socket_, boost::asio::buffer(commandString("getData",COMMANDLENGTH)));
    char  inboundCommand_[COMMANDLENGTH];
    boost::asio::read(socket_, boost::asio::buffer(inboundCommand_));
    string inboundCommand=boost::algorithm::trim_copy(std::string(inboundCommand_, COMMANDLENGTH));
    if(inboundCommand != "compute") {
      socket_.close();
      return string("error"); 
    }
    char inboundHeader_[COMMANDLENGTH];
    boost::asio::read(socket_, boost::asio::buffer(inboundHeader_));
    string inboundHeader=boost::algorithm::trim_copy(std::string(inboundHeader_, COMMANDLENGTH));
    unsigned int dataSize = lexical_cast<unsigned int>(inboundHeader);
    vector<char> inboundData(dataSize);
    boost::asio::read(socket_, boost::asio::buffer(inboundData));
    ostringstream oss;
    vector<char>::iterator it;
    for(it=inboundData.begin(); it!=inboundData.end(); ++it) oss << *it;
    socket_.close();
    return oss.str();
  }

  void submit(string data){
    std::vector<boost::asio::const_buffer> buffers; 
    string result = commandString("result", COMMANDLENGTH);
    buffers.push_back(boost::asio::buffer(result));
    string dataSize = commandString(lexical_cast<string>(data.size()), COMMANDLENGTH);
    buffers.push_back(boost::asio::buffer(dataSize));
    buffers.push_back(boost::asio::buffer(data));
    boost::asio::write(socket_, buffers);
    socket_.close();
  }

  boost::asio::io_service io_service_;
  boost::asio::ip::tcp::socket socket_;

  boost::asio::ip::tcp::resolver resolver_;
  boost::asio::ip::tcp::resolver::query query_;

  boost::asio::ip::tcp::resolver::iterator endpoint_iterator0;
  boost::asio::ip::tcp::resolver::iterator end;
};

/******************************************************************************************/

class session
{
public:
  session(boost::asio::io_service& io_service)
    : socket_(io_service)
  { /* nothing */ }

  tcp::socket& socket() { return socket_; }

  void processRequest() {
    // First check what the client wants from us. This will remove the command from the stream.
    char  inboundCommand_[COMMANDLENGTH];
    boost::asio::read(socket_, boost::asio::buffer(inboundCommand_));
    string inboundCommand=boost::algorithm::trim_copy(std::string(inboundCommand_, COMMANDLENGTH));

    // We accept only two commands: "getData" and "result"
    if(inboundCommand == "getData"){
      // Create a new secretContainer object with VECTORSIZE random numbers,
      // serialize it and send it off
      cout << "Received getData command" << endl;
      secretContainer sc(VECTORSIZE);
      submit(sc.toString());
      cout << "Sent off data" << endl;
    }
    else if(inboundCommand == "result"){
      // De-serialize the data and print the resulting object
      cout << "De-serializing data" << endl;
      string data=retrieve();
      secretContainer sc;
      sc.fromString(data);
      cout << "Received a response" << endl;
    }
    else{
      cout << "Error: unknown command: " << inboundCommand << endl;
      return;
    }
  }

private:
  void submit(string data){
    // The vector holds buffers for a message, the data size and the actual data
    std::vector<boost::asio::const_buffer> buffers; 
    
    // First the message to the server
    string compute = commandString("compute", COMMANDLENGTH);
    buffers.push_back(boost::asio::buffer(compute));
    
    // Now the data size
    string dataSize = commandString(lexical_cast<string>(data.size()), COMMANDLENGTH);
    buffers.push_back(boost::asio::buffer(dataSize));
    
    // And finally the actual data
    buffers.push_back(boost::asio::buffer(data));
    
    // Finally we send the buffers off to the server
    boost::asio::write(socket_, buffers);
  }

  string retrieve(void){
    // Now we extract the size of the data section
    char inboundHeader_[COMMANDLENGTH];
    boost::asio::read(socket_, boost::asio::buffer(inboundHeader_));
    string inboundHeader=boost::algorithm::trim_copy(std::string(inboundHeader_, COMMANDLENGTH));
    unsigned int dataSize = lexical_cast<unsigned int>(inboundHeader);
    
    // Extract the actual data
    vector<char> inboundData(dataSize);
    boost::asio::read(socket_, boost::asio::buffer(inboundData));
    
    // Transfer data into a string
    ostringstream oss;
    vector<char>::iterator it;
    for(it=inboundData.begin(); it!=inboundData.end(); ++it) oss << *it;
    
    // And return the result
    return oss.str();
  }
  
  tcp::socket socket_;
};

/******************************************************************************************/

class server{
public:
  server(string port)
    : acceptor_(io_service_, tcp::endpoint(tcp::v4(), lexical_cast<short>(port))),
      tp(4)
  {
    shared_ptr<session> new_session(new session(io_service_));
    acceptor_.async_accept(new_session->socket(),
			   boost::bind(&server::handle_accept, this, new_session,
				       boost::asio::placeholders::error));
  }

  void handle_accept(shared_ptr<session> current_session, const boost::system::error_code& error) {
    if (error) return;

    // tp.schedule(boost::bind(&session::processRequest,current_session));

    // Make sure a new session is started before we handle the current request
    shared_ptr<session> new_session(new session(io_service_));
    acceptor_.async_accept(new_session->socket(),
			   boost::bind(&server::handle_accept, this, new_session,
				       boost::asio::placeholders::error));

    current_session->processRequest();      
  }

  void run(void){
    io_service_.run();
  }

private:
  boost::asio::io_service io_service_;
  tcp::acceptor acceptor_;
  boost::threadpool::pool tp;
};

/******************************************************************************************/

int main(int argc, char* argv[]){
  try {
    // Are we a client or a server ?
    if(argv[1] == string("client")){      
      if (argc != 4) {
	std::cerr << "Usage: vectorExample client <host> <port>" << endl;
	return 1;
      }
      
      client Client(argv[2], argv[3]);
      Client.run();
    }
    else{
      if (argc != 3) {
	std::cerr << "Usage: vectorExample server <port>" << endl;
	return 1;
      }
      
      server Server(argv[2]);
      Server.run();
    }
  }
  catch (std::exception& e) {
    std::cerr << "Exception: " << e.what() << endl;
  }

  return 0;
}

/******************************************************************************************/
