/*
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Copyright (C) 2007 Dr. Ruediger Berlich - ruediger@berlich.de
 */

// Compile with "g++ -lboost_serialization-gcc42-mt -o serialization1 serialization1.cpp"
// or "g++ -DXMLARCHIVE -lboost_serialization-gcc42-mt -o serialization1 serialization1.cpp"
// In the latter case serialization will be done in XML mode.
// Tested on OpenSUSE 10.3, g++ 4.21, boost 1.34.1 . Note that, by
// default, OpenSUSE 10.3 still comes with boost 1.33.1 . In order to 
// compile Boost 1.34.1 on OpenSUSE 10.3, you might have to add the 
// compiler switch  -fno-strict-aliasing in line 155 of the file 
// $BOOSTHOME/tools/jam/src/build.jam .

#include <iostream>
#include <string>
#include <vector>
#include <string>
#include <sstream>
#include <ctime>

#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/random.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/serialization/utility.hpp>

using namespace std;
using namespace boost;

/******************************************************************************************/

class secretNumber
{
  ///////////////////////////////////////////////////////////////
  friend class boost::serialization::access;
  
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version){
    using boost::serialization::make_nvp;
    ar & make_nvp("secret_", secret_);
  }
  ///////////////////////////////////////////////////////////////
  
public:
  secretNumber(void) : secret_(0) { /* nothing */ }
  secretNumber(double secret) : secret_(secret) { /* nothing */ }
  double value(void) const { return secret_; }

private:
  double secret_;
};

/******************************************************************************************/

template <class T>
class base
  :public vector<T>
{
  ///////////////////////////////////////////////////////////////
  friend class boost::serialization::access;

  template<class Archive>
  void serialize(Archive & ar, const unsigned int version){
    using boost::serialization::make_nvp;
    ar & make_nvp("stdvector", boost::serialization::base_object<vector<T> >(*this));
  }
  ///////////////////////////////////////////////////////////////

public:
  virtual void sort(void) = 0;
};

namespace boost {
  namespace serialization {
    template<class T>
    struct is_abstract<base<T> > {
      typedef mpl::bool_<true> type;
      BOOST_STATIC_CONSTANT(bool, value = true);
    };
}}

/******************************************************************************************/

class secretContainer
  :public base<shared_ptr<secretNumber> >
{
  ///////////////////////////////////////////////////////////////
  friend class boost::serialization::access;
  
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version){
    using boost::serialization::make_nvp;
    ar & make_nvp("base", boost::serialization::base_object<base<shared_ptr<secretNumber> > >(*this));
  }
  ///////////////////////////////////////////////////////////////

public:
  // Initialize random number generators only
  secretContainer(void)
    :mersenne(static_cast<unsigned int> (time(0))),
     normalDist(0.,1.), gauss(mersenne,normalDist)
  { /* nothing */ }

  // Initialize container with random numbers
  secretContainer(unsigned int size)
    :mersenne(static_cast<unsigned int> (time(0))),
     normalDist(0.,1.), gauss(mersenne,normalDist)
  {
    for(unsigned int i=0; i<size; i++){ 
      shared_ptr<secretNumber> p(new secretNumber(gauss()));
      this->push_back(p);
    }
  }

  string toString(void){
    ostringstream oss;  // serialize
    {
#ifdef XMLARCHIVE
      boost::archive::xml_oarchive oa(oss);
#else
      boost::archive::text_oarchive oa(oss);
#endif
      oa << boost::serialization::make_nvp("secretContainer",*this);
    } // archive and stream closed at end of scope    

    return oss.str();
  }

  void fromString(const string& descr){    
    istringstream iss(descr); // de-serialize
    {
#ifdef XMLARCHIVE
      boost::archive::xml_iarchive ia(iss);
#else
      boost::archive::text_iarchive ia(iss);
#endif
      ia >> boost::serialization::make_nvp("secretContainer",*this);
    }
  }

  virtual void sort(void){
    std::sort(this->begin(), this->end(), 
	      bind(&secretNumber::value,_1) < bind(&secretNumber::value,_2));
  }

private:
  mt19937 mersenne;
  normal_distribution<double> normalDist;
  variate_generator<mt19937&, normal_distribution<double> > gauss;
};

/******************************************************************************************/

ostream& operator<< (ostream & s, const secretContainer& sc){
  vector<shared_ptr<secretNumber> >::const_iterator it;
  s.precision(2);
  for(it=sc.begin(); it!=sc.end(); ++it) s << fixed << (*it)->value() << " ";
  return s;
}

/****************************************************************/

int main(int argc, char **argv){
  secretContainer sc(10);
  sc.sort();

  string serialization = sc.toString();

  // Let the user know
  cout << "Printing serialized secretContainer:" << endl;
  cout << serialization << endl;

  // Load serialized object into new secretContainer
  secretContainer sc2;
  sc2.fromString(serialization);

  cout << "Printing content of de-serialized secretContainer:" << endl
       << sc2 << endl;

  return 0;
}

/****************************************************************/
